from .patches import Patches2D
from .loader import TileLoader
from .factory import make_image_patches, make_transcript_patches
from .cluster import cluster_embeddings
from .infer import compute_embeddings
